package src.icc.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import sailpoint.tools.GeneralException;
import sailpoint.tools.Util;
import sailpoint.tools.xml.AbstractXmlObject;


/**
 * Util Class used to log custom method calls and custom scripts.
 * 
 * For Scripts use icc_RuleLibrary_LoggingLibrary (which is calling this class) as a RuleLibrary for your XML Object 
 *
 * Logmessage Format depending on the enabled loglevel:<br>
 * <b>Info</b>: Prefix (Entering, Exiting, Exception)<br>
 * <b>Debug</b>: Prefix and parameters/return value as String<br>
 * <b>Trace</b>: Prefix and parameters/return value as XML Objects
 * (SailpoibtObject) or String<br>
 * 
 * @author Felix Hill iC Consult
 */
public class LoggingUtil {
	
	public static final String PREFIX_METHOD_START = "Entering...";
	public static final String PREFIX_METHOD_END = "Exiting...";
	public static final String PREFIX_METHOD_RETURN_VALUE = "returnValue = ";
	public static final String MAX_RECURSION_STRING = "...(maximum recursion level reached)";
	
	public static final int MAX_DEPTH = 10;

	/**
	 * Method to trace the start of custom java methods including
	 * parameters.<br>
	 * See class javadoc for format and logger names.
	 * 
	 * @author Felix Hill iC Consult
	 * @param logger
	 * @param variables
	 */
	public static void logMethodEntry(Logger logger, Object... variables) {
		
		if (logger.isInfoEnabled())
			logger.info(PREFIX_METHOD_START);
		
		if (!logger.isDebugEnabled() || null == variables)
			return;
		
		saveLogmessage(logger, getMethodParameterString(logger, variables));
	}
	
	/**
	 * Method to trace the end of custom java methods including
	 * parameters and return value.<br>
	 * See class javadoc for format and logger names.
	 * 
	 * @author Felix Hill iC Consult
	 * @param logger
	 * @param returnValue
	 */
	public static void logMethodExit(Logger logger, Object returnValue, Object... variables) {
		
		if (logger.isInfoEnabled())
			logger.info(PREFIX_METHOD_END);
		
		
		if (!logger.isDebugEnabled())
			return;
				
		saveLogmessage(logger, getMethodParameterString(logger, variables) + "\n" + getMethodResultString(logger, returnValue));
	}
	
	/**
	 * Method to log the given loggingString as error
	 * 
	 * @author Christopher Wessling iC Consult
	 * @param logger
	 * @param loggingString
	 */

	public static void logMethodError(Logger logger, String loggingString) {
		logger.error(loggingString);
	}	
	
	/**
	 * Method to log the given loggingString as warn
	 * 
	 * @author Christopher Wessling iC Consult
	 * @param logger
	 * @param loggingString
	 */
	public static void logMethodWarn(Logger logger, String loggingString) {
		logger.warn(loggingString);
	}	
	/**
	 * Returns a String representation of the input Object value.<br>
	 * The format depends on the logger setting if trace is enabled.
	 * <p>
	 * Collections and maps are iterated to a depth of up to 10.<br>
	 * Their values are converted with this method as well.
	 * 
	 * @author Felix Hill iC Consult
	 * @param value
	 * @param traceEnabled
	 * @param depth
	 * @return value as String representation
	 */
	public static String getParameterValue(Object value, boolean traceEnabled, int depth) {
		if (null == value)
			return "null";

		if (depth > MAX_DEPTH)
			return MAX_RECURSION_STRING;

		// print Map
		if (value instanceof Map) {
			Map<String, Object> valueMap = Util.otom(value);
			Map<String, String> printMap = new LinkedHashMap<String, String>();

			for (String key : Util.mapKeys(valueMap))
				printMap.put(key, getParameterValue(valueMap.get(key), traceEnabled, depth + 1));

			return printMap.toString();
		}
		// print List
		else if (value instanceof Collection) {
			Collection<?> valueCollection = (Collection<?>) value;
			ArrayList<String> printList = new ArrayList<String>();

			for (Object listValue : Util.safeIterable(valueCollection))
				printList.add(getParameterValue(listValue, traceEnabled, depth + 1));

			return printList.toString();
		}
		// print SailPointObject if trace is enabled
		else if (traceEnabled && value instanceof AbstractXmlObject) {
			AbstractXmlObject valueSailpointObject = (AbstractXmlObject) value;

			try {
				return valueSailpointObject.toXml();
			} catch (Exception e) {
				return valueSailpointObject.toString();
			}
		}
		// print rest as String
		else
			return String.valueOf(value);
	}
	
	/**
	 * Returns a String representation of the method parameters.<br>
	 * The format depends on the logger setting if trace is enabled.
	 * 
	 * @author Felix Hill iC Consult
	 * @param logger
	 * @param variables
	 * @return returns the method parameters as as String representation
	 */
	public static String getMethodParameterString(Logger logger, Object[] variables) {
		// return if logger debug or lower is disabled
		if (!logger.isDebugEnabled())
			return "";

		return getParameterValues(variables, logger.isTraceEnabled());
	}
	
	/**
	 * Returns a String representation of the method parameters.<br>
	 * The format depends on the logger setting if trace is enabled.
	 * 
	 * @author Felix Hill iC Consult
	 * @param variables
	 * @param traceEnabled  true if trace logging is enabled
	 * @return returns the method parameters as as String representation
	 */
	private static String getParameterValues(Object[] variables, boolean traceEnabled) {
		if (null == variables)
			return "";
		
		String loggingString = "";
		
		// iterate Parameters
		for (int i = 0; i < variables.length; i++) {
			if (null == variables[i]) {
				continue;
			}
			
			loggingString += variables[i].getClass().getSimpleName();
			loggingString += " : ";
			loggingString += getParameterValue(variables[i], traceEnabled, 0);
			if (i != variables.length - 1)
				loggingString += "\n";
		}

		return loggingString;
	}
	
	/**
	 * Method to log a message with the given logger.<br>
	 * The loglevel is chosen dynamically depending on which loglevel is enabled.
	 * <p>
	 * Loglevels are chosen as follows:<br>
	 * Trace<br>
	 * Debug<br>
	 * Info
	 * 
	 * @author Felix Hill iC Consult
	 * @param logger
	 * @param loggingString
	 */
	public static void saveLogmessage(Logger logger, String loggingString) {
		if (logger.isTraceEnabled())
			logger.trace(loggingString);
		else if (logger.isDebugEnabled())
			logger.debug(loggingString);
		else if (logger.isInfoEnabled())
			logger.info(loggingString);
	}
	
	/**
	 * Returns a String representation of the method return value.<br>
	 * The format depends on the enabled logging level, see class javadoc.
	 * 
	 * @author Felix Hill iC Consult
	 * @param logger
	 * @param returnValue
	 * @return the method result as String representation
	 */
	public static String getMethodResultString(Logger logger, Object returnValue) {
		// return if logger debug or lower is disabled
		if (!logger.isDebugEnabled() || null == returnValue)
			return "";

		String loggingString = PREFIX_METHOD_RETURN_VALUE;
		// trace level logging
		if (logger.isTraceEnabled())
			loggingString += getParameterValue(returnValue, true, 0);
		// debug level logging
		else
			loggingString += getParameterValue(returnValue, false, 0);

		return loggingString;
	}
	
	/**
	 * Logs the progress within java methods including parameters.
	 * 
	 * @author Christopher Wessling iC Consult
	 * @param logger
	 * @param loggingString
	 * @param variables
	 */
	public static void logMethodProgress(Logger logger, String loggingString, Object...variables) {
		
		if (!logger.isDebugEnabled())
			return;
		
		if (Util.isEmpty(variables))
			saveLogmessage(logger, loggingString);
		else
			saveLogmessage(logger, loggingString + ": [" + getMethodParameterString(logger, variables) + "]");
	}
	
	/**
	 * Logs the given loggingString and the variables on trace level independently of the current log level
	 * Sets the log level back to the original level after printing the log 
	 * 
	 * @author Christopher Weßling iC Consult
	 * 
	 * @param logger
	 * @param loggingString
	 * @param variables
	 */
	public static void forceDebugLog(Logger logger, String loggingString, Object...variables) {
		Level logLevel = logger.getLevel();
		logger.setLevel(Level.DEBUG);
		logMethodProgress(logger, loggingString, variables);
		logger.setLevel(logLevel);
	}

	/**
	 * Logs the given loggingString and the variables on debug level independently of the current log level
	 * Sets the log level back to the original level after printing the log 
	 * 
	 * @author Christopher Weßling iC Consult
	 * 
	 * @param logger
	 * @param loggingString
	 * @param variables
	 */
	public static void forceTraceLog(Logger logger, String loggingString, Object...variables) {
		Level logLevel = logger.getLevel();
		logger.setLevel(Level.TRACE);
		logMethodProgress(logger, loggingString, variables);
		logger.setLevel(logLevel);
	}
}

